@@ -1,14 +1,14 @@
 from bz2 import BZ2File
-from collections import Counter, namedtuple, UserDict, Sequence, Iterable, \
+from collections import Counter, Sequence, Iterable, \
     Mapping
 from email.mime.text import MIMEText
-from fuzzywuzzy import fuzz, process
 import inspect
 from itertools import chain
 import json
 import os
 from pathlib import Path
 import pickle
+from random import choice
 import re
 import smtplib
 import sys
@@ -16,152 +16,10 @@ import sys
 from htools.config import get_credentials, get_default_user
 
 
-def Args(**kwargs):
-    """Wrapper to easily create a named tuple of arguments. Functions sometimes
-    return multiple values, and we have a few options to handle this: we can
-    return them as a regular tuple, but it is often convenient to be able to
-    reference items by name rather than position. If we want the output to be
-    mutable, we can return a dictionary, but this still requires more space
-    than a tuple and bracket notation is arguably less convenient than dot
-    notation. We can create a new namedtuple inside the function, but this
-    kind of seems like overkill to create a new type of namedtuple for each
-    function.
-
-    Instead, this lets us create a namedtuple of Args on the fly just as easily
-    as making a dictionary.
-
-    Parameters
-    ----------
-
-    Examples
-    --------
-    def math_summary(x, y):
-        sum_ = x + y
-        prod = x * y
-        diff = x - y
-        quotient = x / y
-        return Args(sum=sum_,
-                    product=prod,
-                    difference=diff,
-                    quotient=quotient)
-
-    >>> results = math_summary(4, 2)
-    >>> results.product
-
-    8
-
-    >>> results.quotient
-
-    2
-
-    >>> results
-
-    Args(sum=6, product=8, difference=2, quotient=2)
-    """
-    args = namedtuple('Args', kwargs.keys())
-    return args(*kwargs.values())
-
-
 class InvalidArgumentError(Exception):
     pass
 
 
-class FuzzyKeyDict(dict):
-    """Dictionary that will try to find similar keys if a key is missing and
-    return their corresponding values. This could be useful when working with
-    embeddings, where we could try mapping missing words to a combination of
-    existing words.
-
-    Examples
-    --------
-    d = FuzzyKeyDict(limit=3, verbose=True)
-    d['dog'] = 0
-    d['cat'] = 1
-    d['alley cat'] = 2
-    d['pig'] = 3
-    d['cow'] = 4
-    d['cowbell'] = 5
-    d['baby cow'] = 6
-
-    # Keys and similarity scores for the most similar keys.
-    >>> d.similar_keys('house cat')
-    [('alley cat', 56), ('cat', 50), ('cowbell', 25)]
-
-    # "house cat" not in dict so we get the values for the most similar keys.
-    >>> d['house cat']
-    [2, 1, 5]
-
-    # "cat" is in dict so output is an integer rather than a list.
-    >>> d['cat']
-    1
-    """
-
-    def __init__(self, data=None, limit=3):
-        """
-        Parameters
-        ----------
-        data: Iterable (optional)
-            Sequence of pairs, such as a dictionary or a list of tuples. If
-            provided, this will be used to populate the FuzzyKeyDict.
-        limit: int
-            Number of similar keys to find when trying to retrieve the value
-            for a missing key.
-        """
-        if isinstance(data, Mapping):
-            for k, v in data.items():
-                self[k] = v
-        elif isinstance(data, Iterable):
-            for k, v in data:
-                self[k] = v
-        self.limit = limit
-
-    def __getitem__(self, key):
-        """
-        Returns
-        -------
-        any or list[any]: If key is present in dict, the corresponding value
-            is returned. If not, the n closest keys are identified and their
-            corresponding values are returned in a list (where n is defined
-            by the `limit` argument specified in the constructor). Values are
-            sorted in descending order by the neighboring keys' similarity to
-            the missing key in.
-        """
-        try:
-            return super().__getitem__(key)
-        except KeyError:
-            return [self[k] for k in self.similar_keys(key)]
-
-    def similar_keys(self, key, return_distances=False):
-        pairs = process.extract(key, self.keys(), limit=self.limit,
-                                scorer=fuzz.ratio)
-        if return_distances:
-            return pairs
-        return [p[0] for p in pairs]
-
-
-class LambdaDict(UserDict):
-    """Create a default dict where the default function can accept parameters.
-    Whereas the defaultdict in Collections can set the default as int or list,
-    here we can pass in any function where the key is the parameter.
-    """
-
-    def __init__(self, default_function):
-        """
-        Parameters
-        ----------
-        default_function: function
-            When referencing a key in a LambdaDict object that has not been
-            added yet, the value will be the output of this function called
-            with the key passed in as an argument.
-        """
-        super().__init__()
-        self.f = default_function
-
-    def __missing__(self, key):
-        self[key] = self.f(key)
-        return self[key]
-
-
 def hdir(obj, magics=False, internals=False):
     """Print object methods and attributes, by default excluding magic methods.
 
@@ -222,26 +80,29 @@ def tdir(obj, **kwargs):
     dict[str, type]: Dictionary mapping the name of the object's attributes to
     the corresponding types of those attributes.
     """
-    return {k: type(getattr(obj, k)) 
+    return {k: type(getattr(obj, k))
             for k, v in hdir(obj, **kwargs).items() if v == 'attribute'}
 
 
 def hasarg(func, arg):
-    """Checks if a function has a given argument.
-    Works with args and kwargs as well if you exclude the
-    stars. See example below.
+    """Checks if a function has a given argument. Works with args and kwargs as
+    well if you exclude the stars. See example below.
+
     Parameters
     ----------
     func: function
     arg: str
         Name of argument to look for.
+
     Returns
     -------
     bool
+
     Example
     -------
     def foo(a, b=6, *args):
         return
+
     >>> hasarg(foo, 'b')
     True
     >>> hasarg(foo, 'args')
@@ -418,37 +279,39 @@ def eprint(arr, indent=2, spacing=1):
 
 def _read_write_args(path, mode):
     """Helper for `save` and `load` functions.
-    
+
     Parameters
     ----------
     path: str
         Path to read/write object from/to.
     mode: str
-        'w' for writing files (as in `save`), 'r' for reading files 
+        'w' for writing files (as in `save`), 'r' for reading files
         (as in `load`).
-    
+
     Returns
     -------
     tuple: Function to open file, mode to open file with (str), object to open
         file with.
     """
     ext = path.rpartition('.')[-1]
-    if ext not in {'json', 'pkl', 'zip'}:
+    if ext not in {'json', 'pkl', 'txt', 'zip'}:
         raise InvalidArgumentError(
-            'Invalid extension. Make sure your filename ends with .json, '
-            '.pkl, or .zip.'
+            'Invalid extension. Make sure your filename ends with '
+            '.json, .pkl, or .zip.'
         )
-        
+
     # Store in dict to make it easier to add additional formats in future.
-    ext2data = {'pkl': (open, 'b', pickle), 
-                'zip': (BZ2File, '', pickle), 
-                'json': (open, '', json)}
+    ext2data = {
+        'json': (open, '', json),
+        'pkl': (open, 'b', pickle),
+        'zip': (BZ2File, '', pickle),
+    }
     opener, mode_suffix, saver = ext2data[ext]
     return opener, mode + mode_suffix, saver
 
 
-def save(obj, path, verbose=True):
-    """Wrapper to save data as pickle (optionally zipped) or json.
+def save(obj, path, mode_pre='w', verbose=True):
+    """Wrapper to save data as text, pickle (optionally zipped), or json.
 
     Parameters
     -----------
@@ -456,9 +319,11 @@ def save(obj, path, verbose=True):
         Object to save. This will be pickled/jsonified/zipped inside the
         function - do not convert it before-hand.
     path: str
-        File name to save object to. Should end with .pkl, .zip, or
+        File name to save object to. Should end with .txt, .pkl, .zip, or
         .json depending on desired output format. If .zip is used, object will
         be zipped and then pickled.
+    mode_pre: str
+        Determines whether to write or append text. One of ('w', 'a').
     verbose: bool
         If True, print a message confirming that the data was pickled, along
         with its path.
@@ -471,23 +336,25 @@ def save(obj, path, verbose=True):
     os.makedirs(path.parent, exist_ok=True)
     if verbose: print(f'Writing data to {path}.')
     if path.suffix == '.txt':
-        path.write_text(obj)
+        with path.open(mode_pre) as f:
+            f.write(obj)
     else:
-        opener, mode, saver = _read_write_args(str(path), 'w')
+        opener, mode, saver = _read_write_args(str(path), mode_pre)
         with opener(path, mode) as f:
             saver.dump(obj, f)
 
 
 def load(path, verbose=True):
-    """Wrapper to load pickled (optionally zipped) or json data.
-    
+    """Wrapper to load text files or pickled (optionally zipped) or json data.
+
     Parameters
     ----------
     path : str
-        File to load. File type will be inferred from extension.
+        File to load. File type will be inferred from extension. Must be one of
+        '.txt', '.json', '.pkl', or '.zip'.
     verbose : bool, optional
         If True, will print message stating where object was loaded from.
-    
+
     Returns
     -------
     object: The Python object that was pickled to the specified file.
@@ -505,15 +372,15 @@ def load(path, verbose=True):
 
 def dict_sum(*args):
     """Given two or more dictionaries with numeric values, combine them into a
-     single dictionary. For keys that appear in multiple dictionaries, their
-     corresponding values are added to produce the new value.
+    single dictionary. For keys that appear in multiple dictionaries, their
+    corresponding values are added to produce the new value.
 
-     This differs from combining two dictionaries in the following manner:
+    This differs from combining two dictionaries in the following manner:
 
-     {**d1, **d2}
+    {**d1, **d2}
 
-     The method shown above will combine the keys but will retain the value
-     from d2, rather than adding the values from d1 and d2.
+    The method shown above will combine the keys but will retain the value
+    from d2, rather than adding the values from d1 and d2.
 
     Parameters
     -----------
@@ -750,11 +617,13 @@ def safe_map(func, seq):
 
 
 def flatten(nested):
-    """Flatten a nested sequence where the sub-items can be sequences or 
+    """Flatten a nested sequence where the sub-items can be sequences or
     primitives. This differs slightly from itertools chain methods because
     those require all sub-items to be sequences. Here, items can be primitives,
-    sequences, nested sequences, or any combination of these. This also returns
-    a list rather than a generator.
+    sequences, nested sequences, or any combination of these. Any iterable
+    items aside from strings will be completely un-nested, so use with caution
+    (e.g. a torch Dataset would be unpacked into separate items for each
+    index). This also returns a list rather than a generator.
 
     Parameters
     ----------
@@ -767,9 +636,9 @@ def flatten(nested):
     """
     def _walk(nested):
         for group in nested:
-            try:
+            if isinstance(group, Iterable) and not isinstance(group, str):
                 yield from _walk(group)
-            except TypeError:
+            else:
                 yield group
     return list(_walk(nested))
 
@@ -875,4 +744,100 @@ def vcounts(arr, normalize=True):
     return counts
 
 
+def item(it, random=True, try_values=True):
+    """Get an item from an iterable (e.g. dict, set, torch DataLoader).
+    This is a quick way to access an item for iterables that don't support
+    indexing, or do support indexing but require us to know a key.
+
+    Parameters
+    ----------
+    it: Iterable
+        Container that we want to access a value from.
+    random: bool
+        If True, pick a random value from `it`. Otherwise just return the first
+        value.
+    try_values: bool
+        If True, will check if `it` has a `values` attribute and will operate
+        on that if it does. We often want to see a random value from a dict
+        rather than a key. If we want both a key and value, we could set
+        try_values=False and pass in d.items().
+
+    Returns
+    -------
+    any: An item from the iterable.
+    """
+    if try_values and hasattr(it, 'values'): it = it.values()
+    if random:
+        return choice(list(it))
+    return next(iter(it))
+
+
+def identity(x):
+    """Returns the input argument. Sometimes it is convenient to have this if
+    we sometimes apply a function to an item: rather than defining a None
+    variable, sometimes setting it to a function, then checking if it's None
+    every time we're about to call it, we can set the default as identity and
+    safely call it without checking.
+
+    Parameters
+    ----------
+    x: any
+
+    Returns
+    -------
+    x: Unchanged input.
+    """
+    return x
+
+
+def max_key(d, fn=identity):
+    """Find the maximum value in a dictionary and return the associated key.
+    If we want to compare values using something other than their numeric
+    values, we can specify a function. For example, with a dict mapping strings
+    to strings, fn=len would return the key with the longest value.
+
+    Parameters
+    ----------
+    d: dict
+        Values to select from.
+    fn: callable
+        Takes 1 argument (a single value from d.values()) and returns a number.
+        This will be used to sort the items.
+
+    Returns
+    -------
+    A key from dict `d`.
+    """
+    return max(d.items(), key=lambda x: fn(x[1]))[0]
+
+
+def cd_root(root_subdir='notebooks'):
+    """Run at start of Jupyter notebook to enter project root.
+
+    Parameters
+    ----------
+    root_subdir: str
+        Name of a subdirectory contained in the project root directory.
+        If not found in the current working directory, this will move
+        to the parent directory.
+
+    Examples
+    --------
+    Sample file structure (abbreviated):
+    my_project/
+        py/
+            fetch_raw_data.py
+        notebooks/
+            nb01_eda.ipynb
+
+    Running cd_root() from nb01_eda.ipynb will change the working
+    directory from notebooks/ to my_project/, which is typically the
+    same directory we'd run scripts in py/ from. This makes converting
+    from notebooks to scripts easier.
+    """
+    if root_subdir not in next(os.walk('.'))[1]:
+        os.chdir('..')
+    print('Current directory:', os.getcwd())
+
+
 SENTINEL = object()
