@@ -1,32 +1,34 @@
 from abc import ABC, abstractmethod
 from contextlib import contextmanager, redirect_stdout
 from copy import copy, deepcopy
-from functools import wraps, partial
+from functools import wraps, partial, update_wrapper
 import inspect
+import io
 import logging
 import os
 from pathlib import Path
 import signal
 import sys
 import time
+import types
 import warnings
 from weakref import WeakSet
 
-from htools import hdir, load, save
+from htools import hdir, load, save, identity
 
 
 class AutoInit:
-    """Mixin class where child class has a long list of init arguments where 
-    the parameter name and the class attribute will be the same. Note that 
+    """Mixin class where child class has a long list of init arguments where
+    the parameter name and the class attribute will be the same. Note that
     *args are not supported in the init method because each attribute that is
     defined in the resulting object must have a name. A variable length list
     of args can still be passed in as a single argument, of course, without the
     use of star unpacking.
 
     This updated version of AutoInit is slightly more user friendly than in V1
-    (no more passing locals() to super()) but also slower and probably requires 
-    more testing (all because of the frame hack in the init method). Note that 
-    usage differs from the AutoInit present in htools<=2.0.0, so this is a  
+    (no more passing locals() to super()) but also slower and probably requires
+    more testing (all because of the frame hack in the init method). Note that
+    usage differs from the AutoInit present in htools<=2.0.0, so this is a
     breaking change.
 
     Examples
@@ -67,11 +69,11 @@ class AutoInit:
         # Calculate how many frames to go back to get child class.
         frame_idx = type(self).__mro__.index(AutoInit)
         attrs = {k: v for k, v in sys._getframe(frame_idx).f_locals.items()
-                  if not k.startswith('__')}
+                 if not k.startswith('__')}
         attrs.pop('self')
         bound = inspect.signature(self.__class__.__init__)\
                        .bind_partial(**attrs)
-        
+
         # Flatten dict so kwargs are not listed as their own argument.
         bound.arguments.update(
             bound.arguments.pop('kwargs', {}).get('kwargs', {})
@@ -86,14 +88,14 @@ class AutoInit:
         return:
 
         child = Child('Henry', 8, 'm', 'brown', 52, 70, 3, 'green')
-        Child(name='Henry', age=8, sex='m', hair='brown', height=52, 
+        Child(name='Henry', age=8, sex='m', hair='brown', height=52,
               weight=70, grade=3, eyes='green')
-        
+
         Returns
         -------
         str
         """
-        fstrs = (f'{k}={repr(getattr(self, k))}' for k in self._init_keys)            
+        fstrs = (f'{k}={repr(getattr(self, k))}' for k in self._init_keys)
         return f'{self.__class__.__name__}({", ".join(fstrs)})'
 
 
@@ -134,13 +136,13 @@ def auto_repr(cls):
     Foo(a=3, b='b')
     """
 
-    def repr_(instance):
+    def _repr(instance):
         args = dict(inspect.signature(instance.__init__).parameters)
         arg_strs = (f'{k}={repr(v)}' for k, v in instance.__dict__.items()
                     if k in args.keys())
         return f'{type(instance).__name__}({", ".join(arg_strs)})'
 
-    cls.__repr__ = repr_
+    cls.__repr__ = _repr
     return cls
 
 
@@ -199,7 +201,6 @@ class LoggerMixin:
         for handler in handlers:
             handler.setFormatter(formatter)
             logger.addHandler(handler)
-
         return logger
 
 
@@ -236,13 +237,13 @@ class SaveableMixin:
         return load(path)
 
 
-def chain(func):
+def chainmethod(func):
     """Decorator for methods in classes that want to implement
     eager chaining. Chainable methods should be instance methods
-    that return self. All this decorator does is ensure these
-    methods are called on a deep copy of the instance instead
-    of on the instance itself, so that operations are not done
-    in place.
+    that change 1 or more instance attributes and return None. All this
+    decorator does is ensure these methods are called on a deep copy of the
+    instance instead of on the instance itself so that operations don't affect
+    the original object. The new object is returned.
 
     Examples
     --------
@@ -253,20 +254,17 @@ def chain(func):
             self.arr = arr
             self.b = b
 
-        @chain
+        @chainmethod
         def double(self):
             self.b *= 2
-            return self
 
-        @chain
+        @chainmethod
         def add(self, n):
             self.arr = [x+n for x in self.arr]
-            return self
 
-        @chain
+        @chainmethod
         def append(self, n):
             self.arr.append(n)
-            return self
 
     >>> ec = EagerChainable([1, 3, 5, -22], b=17)
     >>> ec
@@ -283,7 +281,9 @@ def chain(func):
     """
     @wraps(func)
     def wrapper(instance, *args, **kwargs):
-        return func(deepcopy(instance), *args, **kwargs)
+        new_inst = deepcopy(instance)
+        func(new_inst, *args, **kwargs)
+        return new_inst
     return wrapper
 
 
@@ -292,11 +292,9 @@ def lazychain(func):
     LazyChainable class.
     """
     func._is_chainable = True
-
     @wraps(func)
     def wrapped(*args, **kwargs):
         return func(*args, **kwargs)
-
     return wrapped
 
 
@@ -341,7 +339,7 @@ class LazyChainable(metaclass=LazyChainMeta):
     similar to a Spark RDD.
 
     Chainable methods must be decorated with @staticmethod
-    and @chain and be named with a leading underscore. A public
+    and @chainmethod and be named with a leading underscore. A public
     method without the leading underscore will be created, so don't
     overwrite this with another method. Chainable methods
     accept an instance of the same class as the first argument,
@@ -421,6 +419,103 @@ class LazyChainable(metaclass=LazyChainMeta):
         return f'{type(self).__name__}({", ".join(argstrs)})'
 
 
+def params(func):
+    """Get parameters in a functions signature.
+
+    Parameters
+    ----------
+    func: function
+
+    Returns
+    -------
+    dict: Maps name (str) to Parameter.
+    """
+    return dict(inspect.signature(func).parameters)
+
+
+def hasarg(func, arg):
+    """Check if a function has a parameter with a given name. (Technically,
+    hasparam might be a more appropriate name but hasarg lets us match the
+    no-space convention of hasattr and getattr while maintaining readability.)
+
+    Parameters
+    ----------
+    func: function
+    arg: str
+        The name of the parameter that you want to check for in func's
+        signature.
+
+    Returns
+    -------
+    bool: True if `func` has a parameter named `arg`.
+    """
+    return arg in params(func)
+
+
+def bound_args(func, args, kwargs, collapse_kwargs=True):
+    """Get the bound arguments for a function (with defaults applied). This is
+    very commonly used when building decorators that log, check, or alter how
+    a function was called.
+
+    Parameters
+    ----------
+    func: function
+    args: tuple
+        Notice this is not *args. Just pass in the tuple.
+    kwargs: dict
+        Notice this is not **kwargs. just pass in the dict.
+    collapse_kwargs: bool
+        If True, collapse kwargs into the regular parameter dict. E.g.
+        {'a': 1, 'b': True, 'kwargs': {'c': 'c_val', 'd': 0}} ->
+        {'a': 1, 'b': True, 'c': 'c_val', 'd': 0}
+
+    Returns
+    -------
+    OrderedDict[str, any]: Maps parameter name to passed value.
+    """
+    bound = inspect.signature(func).bind_partial(*args, **kwargs)
+    bound.apply_defaults()
+    args = bound.arguments
+    if not collapse_kwargs: return args
+    args.update(args.pop('kwargs', {}))
+    return args
+
+
+def handle_interrupt(func=None, cbs=(), verbose=True):
+    """Decorator that allows us to interrupt a function with ctrl-c. We can
+    pass in callbacks that execute on function end. Keep in mind that local
+    variables will be lost as soon as `func` stops running. If `func` is a
+    method, it may be appropriate to update instance variables while running,
+    which we can access because the instance will be the first element of
+    `args` (passed in as `self`).
+
+    Note: Kwargs are passed to callbacks as a single dict, not as **kwargs.
+
+    Parameters
+    ----------
+    func: function
+    cbs: Iterable[Callback]
+        List of callbacks to execute when `func` completes. These will execute
+        whether we interrupt or not.
+    verbose: bool
+        If True, print a message to stdout when an interrupt occurs.
+    """
+    if not func: return partial(handle_interrupt, cbs=cbs, verbose=verbose)
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        func_inputs = bound_args(func, args, kwargs, collapse_kwargs=False)
+        try:
+            res = func(*args, **kwargs)
+        except KeyboardInterrupt:
+            if verbose: print('KeyboardInterrupt. Aborting...')
+            res = None
+        finally:
+            for cb in cbs:
+                cb.on_end(func, func_inputs, res)
+        return res
+    return wrapper
+
+
 @contextmanager
 def assert_raises(error):
     """Context manager to assert that an error is raised. This can be nice
@@ -441,7 +536,7 @@ def assert_raises(error):
     >>> with assert_raises(ValueError) as ar:
     >>>     a = 'b' + 6
 
-    AssertionError: Wrong error raised. Expected PermissionError, got 
+    AssertionError: Wrong error raised. Expected PermissionError, got
     TypeError(can only concatenate str (not "int") to str)
 
     # Third example throws an error because the code inside the context manager
@@ -481,7 +576,7 @@ def timebox(time, strict=True):
         Max number of seconds before throwing error.
     strict: bool
         If True, timeout will cause an error to be raised, halting execution of
-        the entire program. If False, a warning message will be printed and 
+        the entire program. If False, a warning message will be printed and
         the timeboxed operation will end, letting the program proceed to the
         next step.
 
@@ -519,7 +614,7 @@ def timeboxed(time, strict=True):
         Max number of seconds before throwing error.
     strict: bool
         If True, timeout will cause an error to be raised, halting execution of
-        the entire program. If False, a warning message will be printed and 
+        the entire program. If False, a warning message will be printed and
         the timeboxed operation will end, letting the program proceed to the
         next step.
 
@@ -568,7 +663,7 @@ class cached_property:
      [.4, .13, .06, .55]
      [.77, .14, .05, .9]]
 
-    # Second call accesses attribute without re-computing 
+    # Second call accesses attribute without re-computing
     # (notice no "Building matrix" message).
     >>> v = Vocab(tokens)
     >>> v.embedding_matrix
@@ -583,9 +678,9 @@ class cached_property:
 
     def __get__(self, instance, cls):
         """This method is called when the variable being accessed is not in the
-        instance's state dict. The next time the attribute is accessed, the 
+        instance's state dict. The next time the attribute is accessed, the
         computed value will be in the state dict so this method (and the method
-        in the instance itself) is not called again unless the attribute is 
+        in the instance itself) is not called again unless the attribute is
         deleted.
         """
         # When attribute accessed as class method, instance is None.
@@ -601,9 +696,9 @@ class cached_property:
 
 class ReadOnly:
     """Descriptor to make an attribute read-only. This means that once a value
-    has been set, the user cannot change or delete it. Note that read-only 
-    attributes must first be created as class variables (see example below). 
-    To allow more flexibility, we do allow the user to manually manipulate the 
+    has been set, the user cannot change or delete it. Note that read-only
+    attributes must first be created as class variables (see example below).
+    To allow more flexibility, we do allow the user to manually manipulate the
     instance dictionary.
 
     Examples
@@ -625,12 +720,12 @@ class ReadOnly:
     PermissionError: Attribute is read-only.
 
     >>> del d.breed
-    
+
     PermissionError: Attribute is read-only.
     """
 
     def __init__(self):
-        self.name = None
+        # self.name = None
         self.initialized = WeakSet()
 
     def __set_name__(self, owner, name):
@@ -658,49 +753,49 @@ class ReadOnly:
 
 
 def validating_property(func, allow_del=False):
-    """Factory that makes properties that perform some user-specified 
+    """Factory that makes properties that perform some user-specified
     validation when setting values. The returned function must be used as a
-    descriptor to create a class variable before setting the instance 
+    descriptor to create a class variable before setting the instance
     attribute.
-    
+
     Parameters
     ----------
     func: function
         Function or lambda that accepts a single parameter. This will be used
-        when attempting to set a value for the managed attribute. It should 
+        when attempting to set a value for the managed attribute. It should
         return True if the value is acceptable, False otherwise.
     allow_del: bool
         If True, allow the attribute to be deleted.
-        
+
     Returns
     -------
-    function: A property with validation when setting values. Note that this 
+    function: A property with validation when setting values. Note that this
         will be used as a descriptor, so it must create a class variable as
         shown below. In the example, also notice that the name passed to
         LengthyInt mustt match the name of the variable it is assigned to.
-    
+
     Examples
     --------
     LengthyInt = validating_property(
         lambda x: isinstance(x, int) and len(str(int)) > 4
     )
-    
+
     class Foo:
         long = LengthyInt('long')
         def __init__(self, a, long):
             self.a = a
             self.long = long
-            
+
     >>> foo = Foo(3, 4)
-    
+
     ValueError: Invalid value 4 for argument long.
-   
-    # No error on instantiation because the argument is a valid LengthyInt. 
+
+    # No error on instantiation because the argument is a valid LengthyInt.
     >>> foo = Foo(3, 543210)
     >>> foo.long
 
     543210
-   
+
     >>> foo = Foo(3, 'abc')
     ValueError: Invalid value 'abc' for argument long.
     """
@@ -708,14 +803,14 @@ def validating_property(func, allow_del=False):
         @property
         def method(instance):
             return instance.__dict__[name]
-        
+
         @method.setter
         def method(instance, val):
             if func(val):
                 instance.__dict__[name] = val
             else:
                 raise ValueError(f'Invalid value {val} for argument {name}.')
-       
+
         if allow_del:
             @method.deleter
             def method(instance):
@@ -736,10 +831,22 @@ class Callback(ABC):
     """
 
     @abstractmethod
-    def on_begin(self, inputs, output=None):
+    def setup(self, func):
         """
         Parameters
-        -------------
+        ----------
+        func: function
+            The function being decorated.
+
+        """
+
+    @abstractmethod
+    def on_begin(self, func, inputs, output=None):
+        """
+        Parameters
+        ----------
+        func: function
+            The function being decorated.
         inputs: dict
             Dictionary of bound arguments passed to the function being
             decorated with @callbacks.
@@ -750,10 +857,12 @@ class Callback(ABC):
         """
 
     @abstractmethod
-    def on_end(self, inputs, output=None):
+    def on_end(self, func, inputs, output=None):
         """
         Parameters
-        -------------
+        ----------
+        func: function
+            The function being decorated.
         inputs: dict
             Dictionary of bound arguments passed to the function being
             decorated with @callbacks.
@@ -786,47 +895,49 @@ def callbacks(cbs):
         # Train model.
     """
     def decorator(func):
+        for cb in cbs:
+            cb.setup(func)
         @wraps(func)
         def wrapper(*args, **kwargs):
             bound = inspect.signature(func).bind_partial(*args, **kwargs)
             bound.apply_defaults()
             for cb in cbs:
-                cb.on_begin(bound.arguments, None)
+                cb.on_begin(func, bound.arguments, None)
             out = func(*args, **kwargs)
             for cb in cbs:
-                cb.on_end(bound.arguments, out)
+                cb.on_end(func, bound.arguments, out)
             return out
         return wrapper
     return decorator
 
 
 def typecheck(func_=None, **types):
-    """Decorator to enforce type checking for a function or method. There are 
+    """Decorator to enforce type checking for a function or method. There are
     two ways to call this: either explicitly passing argument types to the
     decorator, or letting it infer them using type annotations in the function
     that will be decorated. We allow multiple both usage methods since older
-    versions of Python lack type annotations, and also because I feel the 
+    versions of Python lack type annotations, and also because I feel the
     annotation syntax can hurt readability.
 
     Parameters
     ----------
     func_: function
-        The function to decorate. When using decorator with 
+        The function to decorate. When using decorator with
         manually-specified types, this is None. Underscore is used so that
-        `func` can still be used as a valid keyword argument for the wrapped 
+        `func` can still be used as a valid keyword argument for the wrapped
         function.
     types: type
         Optional way to specify variable types. Use standard types rather than
-        importing from the typing library, as subscripted generics are not 
+        importing from the typing library, as subscripted generics are not
         supported (e.g. typing.List[str] will not work; typing.List will but at
-        that point there is no benefit over the standard `list`). 
-        
+        that point there is no benefit over the standard `list`).
+
     Examples
     --------
     In the first example, we specify types directly in the decorator. Notice
-    that they can be single types or tuples of types. You can choose to 
+    that they can be single types or tuples of types. You can choose to
     specify types for all arguments or just a subset.
-    
+
     @typecheck(x=float, y=(int, float), iters=int, verbose=bool)
     def process(x, y, z, iters=5, verbose=True):
         print(f'z = {z}')
@@ -834,19 +945,19 @@ def typecheck(func_=None, **types):
             if verbose: print(f'Iteration {i}...')
             x *= y
         return x
-    
+
     >>> process(3.1, 4.5, 0, 2.0)
     TypeError: iters must be <class 'int'>, not <class 'float'>.
-    
+
     >>> process(3.1, 4, 'a', 1, False)
     z = a
     12.4
-    
+
     Alternatively, you can let the decorator infer types using annotations
-    in the function that is to be decorated. The example below behaves 
+    in the function that is to be decorated. The example below behaves
     equivalently to the explicit example shown above. Note that annotations
     regarding the returned value are ignored.
-    
+
     @typecheck
     def process(x:float, y:(int, float), z, iters:int=5, verbose:bool=True):
         print(f'z = {z}')
@@ -854,10 +965,10 @@ def typecheck(func_=None, **types):
             if verbose: print(f'Iteration {i}...')
             x *= y
         return x
-        
+
     >>> process(3.1, 4.5, 0, 2.0)
     TypeError: iters must be <class 'int'>, not <class 'float'>.
-    
+
     >>> process(3.1, 4, 'a', 1, False)
     z = a
     12.4
@@ -867,10 +978,10 @@ def typecheck(func_=None, **types):
         return partial(typecheck, **types)
     # Case 2: Infer types from annotations. Skip if Case 1 already occurred.
     elif not types:
-        types = {k: v.annotation 
+        types = {k: v.annotation
                  for k, v in inspect.signature(func_).parameters.items()
                  if not v.annotation == inspect._empty}
-    
+
     @wraps(func_)
     def wrapper(*args, **kwargs):
         fargs = inspect.signature(wrapper).bind(*args, **kwargs).arguments
@@ -986,14 +1097,13 @@ def debug(func=None, prefix='', arguments=True):
         out_fmt = '\n{}CALLING {}({})'
         arg_strs = ''
         if arguments:
-            sig = inspect.signature(wrapper).bind_partial(*args, **kwargs)
-            sig.apply_defaults()
-            sig.arguments.update(sig.arguments.pop('kwargs', {}))
-            # Remove self/cls arg from methods.
-            first_key = next(iter(sig.arguments))
-            if first_key in ('self', 'cls'):
-                del sig.arguments[first_key]
-            arg_strs = (f'{k}={repr(v)}' for k, v in sig.arguments.items())
+            sig = bound_args(wrapper, args, kwargs, collapse_kwargs=True)
+            if sig:
+                first_key = next(iter(sig))
+                # Remove self/cls arg from methods.
+                if first_key in ('self', 'cls'):
+                    del sig[first_key]
+            arg_strs = (f'{k}={repr(v)}' for k, v in sig.items())
 
         # Print call message and return output.
         print(out_fmt.format(prefix, func.__qualname__, ', '.join(arg_strs)))
@@ -1045,53 +1155,118 @@ def log_stdout(func=None, fname=''):
     return wrapper
 
 
-def wrapmethods(*decorators, magics=False, internals=False):
+def return_stdout(func):
+    """Decorator that returns printed output from the wrapped function. This
+    may be useful if we define a function that only prints information and
+    returns nothing, then later decide we want to access the printed output.
+    Rather than re-writing everything, we can slap a @return_stdout decorator
+    on top and leave it as is. This should not be used if the decorated
+    function already returns something else since we will only return what is
+    printed to stdout. For that use case, consider the `log_stdout` function.
+    """
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        res = io.StringIO()
+        with redirect_stdout(res):
+            func(*args, **kwargs)
+        return res.getvalue()
+    return wrapper
+
+
+def log_cmd(path, mode='a'):
+    """Decorator that saves the calling command for a python script. This is
+    often useful for CLIs that train ML models. It makes it easy to re-run
+    the script at a later date with the same or similar arguments.
+
+    Parameters
+    ----------
+    path: str or Path
+        Specifies file where output will be saved.
+    mode: str
+        Determines whether output should overwrite old file or be appended.
+        One of ('a', 'w'). In most cases we will want append mode because we're
+        tracking multiple trials.
+
+    Examples
+    --------
+    ```
+    # train.py
+    import fire
+
+    @log_cmd('logs/training_runs.txt')
+    def train(lr, epochs, dropout, arch, data_version):
+        # Train model
+
+    if __name__ == '__main__':
+        fire.Fire(train)
+    ```
+
+    $ python train.py --lr 3e-3 --epochs 50 --dropout 0.5 --arch awd_lstm \
+        --data_version 1
+
+    After running the script with the above command, the file
+    'logs/training_runs.txt' now contains a nicely formatted version of the
+    calling command with a separate line for each argument name/value pair.
+
+    We can also use variables that are passed to our function. All function
+    args and kwargs will be passed to the string formatter so your variable
+    names must match:
+
+    @log_cmd('logs/train_run_v{version_number}.{ext}')
+    def train(version_number, ext, epochs, arch='lstm'):
+        # Train model
+    """
+    def decorator(func):
+        @wraps(func)
+        def wrapped(*args, **kwargs):
+            fn_locals = bound_args(func, args, kwargs, True)
+            res = 'python'
+            for arg in sys.argv:
+                pre = ' \\\n\t' if arg.startswith('-') else ' '
+                res += pre+arg
+            save(res+'\n\n', Path(path.format(**fn_locals)), mode)
+            return func(*args, **kwargs)
+        return wrapped
+    return decorator
+
+
+def wrapmethods(*decorators, methods=(), internals=False):
     """Class wrapper that applies 1 or more decorators to every non-magic
-    method. For example, we often want @debug to be applied to many different
-    methods.
+    method (properties are also excluded). For example, we often want @debug
+    to be applied to many different methods.
 
     Parameters
     ----------
     decorators: callable
         1 or more decorators to apply to methods within a class. By default,
         methods with 1 or 2 leading underscores are excluded.
-    magics: bool
-        If True, apply decorators to methods named with leading double 
-        underscores.
+    methods: Iterable[str]
+        Names of methods to wrap if you don't want to wrap all of them.
+        Internal methods can be wrapped but magic methods and properties
+        cannot.
     internals: bool
-        If True, apply decorators to methods named with leading single 
-        underscores.
-
-    Examples
-    --------
-    @wrapmethods(typecheck, debug)
-    class Foo:
-        def __init__(self, a, b=6):
-            self.a = a
-            self.b = b
-        
-        def walk(self, c:str, d=0, e:bool=True):
-            # This method will enforce type-checking and a message will be
-            # printed to stdout when it is called.
-
-        def run(self, y, z:str):
-            # This method will also have type-checking and debug messages.
-
-        def _jump(self, n:int):
-            # This method will not enforce annotations or print debug messages
-            # because the leading underscore indicates that it is an internal
-            # method. If we did want it to be decorated, we would pass
-            # `internals=True` into the `wrapmethods` decorator call as keyword
-            # arguments.
+        If True, apply decorators to methods named with leading single
+        underscores. This will be ignored if `methods` is specified.
     """
+
     def wrapper(cls):
-        for attr, is_method in hdir(cls, magics, internals).items():
-            if not is_method:
+        special_methods = (staticmethod, classmethod)
+        to_wrap = dict.fromkeys(methods, True) if methods \
+            else hdir(cls, False, internals=internals)
+        for attr, is_method in to_wrap.items():
+            f = cls.__dict__[attr]
+            if not is_method or isinstance(f, property):
                 continue
-            wrapped = getattr(cls, attr)
+
+            # Classmethod and staticmethod decorators need to be applied last.
+            final_wrapper = identity
+            if isinstance(f, special_methods):
+                final_wrapper = type(f)
+                f = f.__func__
+
             for d in decorators:
-                wrapped = d(wrapped)
-                setattr(cls, attr, wrapped)
+                f = d(f)
+            setattr(cls, attr, final_wrapper(f))
         return cls
     return wrapper
 
@@ -1143,7 +1318,6 @@ def delegate(attr):
             return getattr(delegate, new_attr)
         cls.__getattr__ = f
         return cls
-
     return wrapper
 
 
@@ -1218,3 +1392,95 @@ def block_timer():
     finally:
         duration = time.perf_counter() - start
         print(f'[TIMER]: Block executed in {duration:.3f} seconds.')
+
+
+def count_calls(func):
+    """Count the number of times a function has been called. The function can
+    access this value inside itself through the attribute 'calls'. Note that
+    counting is defined such that during the first call, func.calls already=1
+    (i.e. it can be considered the n'th call, not that n calls have previously
+    taken place not counting the current one).
+    """
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        calls = getattr(wrapper, 'calls', 0)
+        wrapper.calls = calls + 1
+        return func(*args, **kwargs)
+    return wrapper
+
+
+def copy_func(func):
+    """Copy a function. Regular copy and deepcopy functionality do not work
+    on functions the way they do on most objects. If we want to create a new
+    function based on another without altering the old one (as in
+    `rename_params`), this should be used.
+
+    Parameters
+    ----------
+    func: function
+        Function to duplicate.
+
+    Returns
+    -------
+    function: Copy of input `func`.
+
+    Examples
+    --------
+    def foo(a, b=3, *args, c=5, **kwargs):
+        return a, b, c, args, kwargs
+
+    foo2 = copy_func(foo)
+
+    >>> foo2.__code__ == foo.__code__
+    True
+
+    >>> foo2 == foo
+    False
+    """
+    new_func = types.FunctionType(func.__code__, func.__globals__,
+                                  func.__name__, func.__defaults__,
+                                  func.__closure__)
+    new_func.__kwdefaults__ = func.__kwdefaults__.copy()
+    return update_wrapper(new_func, func)
+
+
+def rename_params(func, **old2new):
+    """Rename one or more parameters. Docstrings and default arguments are
+    updated accordingly. This is useful when working with code that uses
+    `hasarg`. For example, my Incendio library uses parameter names
+    to pass the correct arguments to different metrics.
+
+    Parameters
+    ----------
+    func: function
+        The old function to change.
+    old2new: str
+        One or more parameter names to change and their corresponding new
+        names. See Example below for a more concrete example.
+
+    Returns
+    -------
+    function: Same as input `func` but with updated parameter names.
+
+    Examples
+    --------
+    def foo(a, b, *args, c=3, **kwargs):
+        pass
+
+    foo_metric = rename_params(func, a=y_true, b=y_pred)
+
+    `foo_metric` will work exactly like `foo` but its first two parameters will
+    now be named "y_true" and "y_pred", respectively. """
+    new_func = copy_func(func)
+    sig = inspect.signature(new_func)
+    kw_defaults = func.__kwdefaults__
+    names, params = map(list, zip(*sig.parameters.items()))
+    for old, new in old2new.items():
+        idx = names.index(old)
+        default = kw_defaults.get(old)
+        kwargs = {'default': default} if default else {}
+        params[idx] = inspect.Parameter(new, params[idx].kind, **kwargs)
+    new_func.__signature__ = sig.replace(parameters=params)
+    return new_func
+
+
